# SPDX-FileCopyrightText: 2026 Oak Ridge National Laboratory and Contributors
#
# SPDX-License-Identifier: Apache-2.0

from os import fstat

import numpy as np


def ReadMetaMetadataRecord(buf, rec, pos, fileSize):
    # Read one metametadata record
    startoffset = str(pos).rjust(8)
    # 8 bytes MetaMetaIDLen + MetaMetaInfoLen Length
    mmIDlen = np.frombuffer(buf, dtype=np.uint64, count=1, offset=pos)
    pos = pos + 8
    mmInfolen = np.frombuffer(buf, dtype=np.uint64, count=1, offset=pos)
    pos = pos + 8

    id_start = pos
    id_bytes = bytes(buf[id_start : id_start + int(mmIDlen[0])])
    pos = pos + int(mmIDlen[0])

    info_start = pos
    info_bytes = bytes(buf[info_start : info_start + int(mmInfolen[0])])
    pos = pos + int(mmInfolen[0])

    recs = str(rec).rjust(7)
    idlen = str(mmIDlen[0]).rjust(10)
    infolen = str(mmInfolen[0]).rjust(12)

    print(f" | {recs} | {startoffset} | {idlen} | {infolen} |")

    return pos, id_bytes, info_bytes


def DumpMetaMetaData(fileName, ffsDecoder=None):
    print("========================================================")
    print("    MetaMetadata File: " + fileName)
    print("========================================================")
    mmd_records = []
    with open(fileName, "rb") as f:
        fileSize = fstat(f.fileno()).st_size
        print(f" File size = {fileSize}")
        buf = f.read(fileSize)
        print(" --------------------------------------------------")
        print(" |  Record |  Offset  |  ID length |  Info length |")
        print(" --------------------------------------------------")
        pos = 0
        rec = 0
        while pos < fileSize - 1:
            pos, id_bytes, info_bytes = ReadMetaMetadataRecord(buf, rec, pos, fileSize)
            mmd_records.append((id_bytes, info_bytes))
            rec = rec + 1
        print(" --------------------------------------------------")

        if ffsDecoder:
            _print_metametadata_fields(mmd_records, ffsDecoder)

    return True, mmd_records


def _print_metametadata_fields(mmd_records, ffsDecoder):
    """Print the decoded field lists from metametadata records."""
    ffsDecoder.load_metametadata(mmd_records)
    print()
    print("  Decoded format definitions:")
    for idx, (id_bytes, info_bytes) in enumerate(mmd_records):
        # Load this single record into a temporary context to get its
        # format info. We already loaded all records above, so we can
        # use FMformat_from_ID.
        fm_format = ffsDecoder.lib.FMformat_from_ID(ffsDecoder.fm_context, id_bytes)
        if not fm_format:
            continue
        format_name = ffsDecoder.lib.name_of_FMformat(fm_format)
        if format_name:
            format_name = format_name.decode("utf-8", errors="replace")
        else:
            format_name = f"<record {idx}>"

        desc_list_ptr = ffsDecoder.lib.format_list_of_FMFormat(fm_format)
        field_list = ffsDecoder._read_field_list(desc_list_ptr)

        print(f'  Format "{format_name}":')
        for name, ftype, fsize, foffset in field_list:
            print(f"    {name:40s} {ftype:30s} size={fsize} offset={foffset}")
        print()


if __name__ == "__main__":
    print("ERROR: Utility main program is bp5dbg.py")



# SPDX-FileCopyrightText: 2026 Oak Ridge National Laboratory and Contributors
#
# SPDX-License-Identifier: Apache-2.0

"""ADIOS2 Stream Python wrapper."""

from functools import singledispatchmethod
from sys import maxsize
from typing import Optional
import numpy as np

try:
    import cupy as cp

    ADIOS2_HAS_CUPY = True
except ImportError:
    ADIOS2_HAS_CUPY = False
try:
    import torch

    ADIOS2_HAS_TORCH = True
except ImportError:
    ADIOS2_HAS_TORCH = False

from adios2 import bindings, Adios, IO, Variable


def type_adios_to_numpy(name):
    """Translation between numpy and adios2 types"""
    return {
        "char": np.int8,
        "int8_t": np.int8,
        "uint8_t": np.uint8,
        "int16_t": np.int16,
        "uint16_t": np.uint16,
        "int32_t": np.int32,
        "uint32_t": np.uint32,
        "int64_t": np.int64,
        "uint64_t": np.uint64,
        "float": np.float32,
        "double": np.float64,
        "complex": np.complex64,
        "double complex": np.complex128,
    }[name]


def string_to_mode(mode: str) -> [bindings.Mode, bool]:
    """Convert high level open mode to adios2.bindings.Mode"""
    read_mode = False
    if mode == "r":
        bmode = bindings.Mode.Read
        read_mode = True
    elif mode == "rra":
        bmode = bindings.Mode.ReadRandomAccess
        read_mode = True
    elif mode == "w":
        bmode = bindings.Mode.Write
    elif mode == "a":
        bmode = bindings.Mode.Append
    else:
        raise ValueError()
    return bmode, read_mode


class Stream:  # noqa: PLR0902
    """High level implementation of the Stream class from the core API"""

    # Default timeout for stream.begin_step()
    DEFAULT_TIMEOUT_SEC = -1.0

    @singledispatchmethod
    def __init__(self, path, mode, comm=None):  # noqa: PLR0912
        if comm and not bindings.is_built_with_mpi:
            raise RuntimeError("Cannot use MPI since ADIOS2 was built without MPI support")

        if comm:
            self._adios = Adios(comm)
        else:
            self._adios = Adios()

        self._io_name = f"stream:{path}:mode:{mode}"

        self._io = self._adios.declare_io(self._io_name)
        self._mode, self._read_mode = string_to_mode(mode)
        self._engine = self._io.open(path, self._mode)
        self.index = -1
        self.max_steps = maxsize
        self._step_status = bindings.StepStatus.EndOfStream
        self._step_timeout_sec = self.DEFAULT_TIMEOUT_SEC

    # e.g. Stream(io: adios2.IO, path, mode)
    @__init__.register(IO)
    def _(self, io: IO, path, mode, comm=None):
        self._io = io
        self._adios = io.adios()
        self._mode, self._read_mode = string_to_mode(mode)
        self._engine = self._io.open(path, self._mode, comm)
        self.index = -1
        self.max_steps = maxsize
        self._step_status = bindings.StepStatus.EndOfStream
        self._step_timeout_sec = self.DEFAULT_TIMEOUT_SEC

    @property
    def mode(self):
        """Selected open mode"""
        return self._mode

    @property
    def adios(self):
        """Adios instance associated to this Stream"""
        return self._adios

    @property
    def io(self):
        """IO instance associated to this Stream"""
        return self._io

    @property
    def engine(self):
        """Engine instance associated to this Stream"""
        return self._engine

    def __repr__(self):
        return f"<adios.file named {self._io_name} and mode {self._mode}>"

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close()

    def __iter__(self):
        return self

    def __next__(self):
        if self.index >= 0 and self._step_status == bindings.StepStatus.OK:
            self.end_step()
        if self.index == self.max_steps - 1:
            self._step_status = bindings.StepStatus.EndOfStream
            raise StopIteration

        self.index += 1
        self._step_status = self.begin_step(timeout=self._step_timeout_sec)
        if self._step_status == bindings.StepStatus.EndOfStream:
            raise StopIteration

        if self._step_status == bindings.StepStatus.NotReady:
            print(
                "ERROR: Stream returned no new step within the time limit of"
                f" {self._step_timeout_sec} seconds. Ending the loop"
            )
            raise StopIteration

        if self._step_status == bindings.StepStatus.OtherError:
            print("ERROR: Stream returned an error. Ending the loop")
            raise StopIteration

        return self

    def get_metadata(self):
        """
        Get metadata of an open file in a serialized form that can be
        sent to other processes and used to open the file again by
        avoiding reading the metadata from disk.
        """
        return self._engine.get_metadata()

    def step_status(self):
        """Inspect the stream status. Return adios2.bindings.StepStatus"""
        return self._step_status

    def set_parameters(self, **kwargs):
        """
        Sets parameters using a dictionary.
        Removes any previous parameter.

        Parameters
            parameters
                input key/value parameters

            value
                parameter value
        """
        self._io.set_parameters(**kwargs)

    def available_variables(self):
        """
        Returns a 2-level dictionary with variable information.
        Read mode only.

        Returns
            variables dictionary
                key
                    variable name
                value
                    variable information dictionary
        """
        return self._io.available_variables()

    def available_attributes(self, varname="", separator="/"):
        """
        Returns a 2-level dictionary with attribute information.
        Read mode only.

        Parameters
            variable_name
                If varname is set, attributes assigned to that variable are returned.
                The keys returned are attribute names with the prefix of varname + separator
                removed.

            separator
                concatenation string between variable_name and attribute
                e.g. varname + separator + name ("var/attr")
                Not used if varname is empty

        Returns
            attributes dictionary
                key
                    attribute name
                value
                    attribute information dictionary
        """
        return self._io.available_attributes(varname, separator)

    def define_variable(self, name):
        """
        Define new variable without specifying its type and content.
        This only works for string output variables
        """
        return self._io.define_variable(name)

    def inquire_variable(self, name):
        """
        Inquire a variable

        Parameters
            name
                variable name
        Returns
            The variable if it is defined, otherwise None
        """
        return self._io.inquire_variable(name)

    def inquire_attribute(self, name, variable_name="", separator="/"):
        """
        Inquire an attribute

        Parameters
            name
                attribute name

            variable_name
                if attribute is associated with a variable

            separator
                concatenation string between variable_name and attribute
                e.g. variable_name + separator + name ("var/attr")
                Not used if variable_name is empty

        Returns
            The attribute if it is defined, otherwise None
        """
        return self._io.inquire_attribute(name, variable_name, separator)

    @singledispatchmethod
    def write(self, variable: Variable, content):
        """
        Writes a variable.
        Note that the content will be available for consumption only at
        the end of the for loop or in the end_step() call.

        Parameters
            variable
                adios2.Variable object to be written
                Use variable.set_selection(), set_shape(), add_operation_string()
                to prepare a write

            content
                variable data values
        """
        self._engine.put(variable, content, bindings.Mode.Sync)

    def _get_variable_selection(self, content, shape, start, count):
        """
        Internal function to get the shape, start and count when defining a variable for write.

        Parameters
           content
                The array that will be written to a variable
            shape
                The shape requested for the write
            start
                The start requested for the write
            count
                The count requested for the write
        Returns
            shapt, start, count
                Either what was selected or dimensions given by the content
        """
        if shape == [] and start == [] and count == []:
            shape = list(content.shape)
            start = [0] * content.ndim
            count = shape[:]
        return shape, start, count

    @write.register(str)
    def _(self, name, content, shape=[], start=[], count=[], operations=None):
        """
        Writes a variable

        Parameters
            name
                variable name

            content
                variable data values

            shape
                variable global MPI dimensions.

            start
                variable offset for current MPI rank.

            count
                variable dimension for current MPI rank.

            operations
                operations to be used in this variable
        """
        variable = self._io.inquire_variable(name)

        if not variable:
            # Sequence variables
            if isinstance(content, (list, np.ndarray)):
                if isinstance(content, list):
                    content = np.asarray(content)

                # If shape, start, and count is not specified, use the numpy array's shape
                shape, start, count = self._get_variable_selection(content, shape, start, count)
                variable = self._io.define_variable(name, content, shape, start, count)
            elif ADIOS2_HAS_CUPY and isinstance(content, cp.ndarray):
                shape, start, count = self._get_variable_selection(content, shape, start, count)
                variable = self._io.define_variable(
                    name, np.array([0], dtype=content.dtype), shape, start, count
                )
            elif ADIOS2_HAS_TORCH and isinstance(content, torch.Tensor):
                shape, start, count = self._get_variable_selection(content, shape, start, count)
                var_type = str(content.dtype).rsplit(".", maxsplit=1)[-1]
                variable = self._io.define_variable(
                    name, np.array([0], dtype=var_type), shape, start, count
                )
            # Scalar variables
            elif isinstance(content, str):
                variable = self._io.define_variable(name, content)
            elif not hasattr(content, "__len__"):
                variable = self._io.define_variable(name, content, [], [], [])
            else:
                raise ValueError

        if shape != [] and not variable.single_value():
            variable.set_shape(shape)

        if start != [] or count != []:
            variable.set_selection([start, count])

        if operations:
            variable.remove_operations()
            for operation in operations:
                variable.add_operation_string(operation[0], operation[1])

        self.write(variable, content)

    def _read_var(self, variable: Variable, defer_read: bool = False):
        """
        Internal function to read when there is no preallocated buffer submitted.
        Settings must be done to Variable before the call.

        Parameters
            variable
                adios2.Variable object to be read
                Use variable.set_selection(), set_block_selection(), set_step_selection()
                to prepare a read
            defer_read
                False: read now and blocking wait for completion (Sync mode)
                True: defer reading all requests until read_complete().
                        The returned numpy array will be filled with data
                        only after calling read_complete().
        Returns
            array
                resulting array from selection
        """
        dtype = type_adios_to_numpy(variable.type())
        count = variable.count()

        if count != []:
            # array
            # steps = variable.get_steps_from_step_selection()
            # if steps == 0:
            #     steps = 1
            # Missing from C++ API: get the steps set for a variable.
            # Calculate the steps from size of count and total size that
            # we can get from the C++ API.
            size_per_step = np.prod(count)
            size_all_steps = variable.selection_size()
            steps = int(size_all_steps / size_per_step)
            if size_all_steps % size_per_step != 0:
                print(
                    f"Stream read(), step calculation for array went horribly wrong "
                    f" variable name = {variable.name()}"
                    f" selection size = {size_all_steps}  size per step = {size_per_step}"
                )

            output_shape = np.array(count)
            output_shape[0] *= steps
        else:
            # scalar
            size_all_steps = variable.selection_size()
            # if size_all_steps > 1:
            if self._mode == bindings.Mode.ReadRandomAccess and variable.steps() > 1:
                output_shape = [size_all_steps]
            else:
                output_shape = []

        mode = bindings.Mode.Sync
        if defer_read:
            mode = bindings.Mode.Deferred

        output = np.zeros(output_shape, dtype=dtype)
        self._engine.get(variable, output, mode)
        return output

    def _set_variable_settings(self, variable, start, count, block_id, step_selection):
        """
        Internal function to set the settings to Variable before getting data.

        Parameters
            variable
                adios2.Variable object to be read

            start
                variable offset dimensions

            count
                variable local dimensions from offset

            block_id
                (int) Required for reading local variables, local array, and local
                value.

            step_selection
                (list): On the form of [start, count].
        Returns
            variable
                the variable with the selection set
        """
        if step_selection is not None and not self._mode == bindings.Mode.ReadRandomAccess:
            raise RuntimeError("step_selection parameter requires 'rra' mode")

        if step_selection is not None:
            variable.set_step_selection(step_selection)

        if block_id is not None:
            variable.set_block_selection(block_id)

        if variable.type() == "string" and variable.single_value() is True:
            return variable

        if start != [] and count != []:
            variable.set_selection([start, count])
        return variable

    @singledispatchmethod
    def read_in_buffer(
        self,
        variable: Variable,
        buffer,
        start=[],
        count=[],
        block_id=None,
        step_selection=None,
        defer_read: bool = False,
    ):
        """
        Read a variable into a preallocated buffer.
        Random access read allowed to select steps.

        Parameters
            variable
                adios2.Variable object to be read
                Use variable.set_selection(), set_block_selection(), set_step_selection()
                to prepare a read

            buffer
                the pre-allocated buffer that will hold the read data
                (numpy, cupy, torch.Tensor arrays)

            start
                variable offset dimensions

            count
                variable local dimensions from offset

            block_id
                (int) Required for reading local variables, local array, and local
                value.

            step_selection
                (list): On the form of [start, count].

            defer_read
                False: read now and blocking wait for completion (Sync mode)
                True: defer reading all requests until read_complete().
                      The returned numpy array will be filled with data
                      only after calling read_complete().
        """
        variable = self._set_variable_settings(variable, start, count, block_id, step_selection)
        # make sure the buffer is a mutable array
        islist = False
        if isinstance(buffer, (list, np.ndarray)):
            islist = True
        elif ADIOS2_HAS_CUPY and isinstance(buffer, cp.ndarray):
            islist = True
        elif ADIOS2_HAS_TORCH and isinstance(buffer, torch.Tensor):
            islist = True
        if not islist:
            raise RuntimeError(
                "Cannot read single value in predefined buffers. "
                "Please use val = stream.read(var, ...) form instead"
            )

        if ADIOS2_HAS_TORCH and isinstance(buffer, torch.Tensor):
            buf_type = str(buffer.dtype).rsplit(".", maxsplit=1)[-1]
        else:
            buf_type = buffer.dtype

        dtype = type_adios_to_numpy(variable.type())
        if dtype(10).dtype != buf_type:
            raise RuntimeError(f"Read buffer type {buf_type} does not match variable type {dtype}")

        count = np.prod(variable.count())
        buf_size = buffer.size
        if ADIOS2_HAS_TORCH and isinstance(buffer, torch.Tensor):
            buf_size = buffer.numel()
        if count != buf_size:
            raise RuntimeError("Read buffer size {buf_size} does not match variable size {count}")

        mode = bindings.Mode.Sync
        if defer_read:
            mode = bindings.Mode.Deferred

        self._engine.get(variable, buffer, mode)

    @read_in_buffer.register(str)
    def _(
        self,
        name: str,
        buffer,
        start=[],
        count=[],
        block_id=None,
        step_selection=None,
        defer_read: bool = False,
    ):
        """
        Read a variable into a preallocated buffer.
        Random access read allowed to select steps.

        Parameters
            name
                variable to be read

            buffer
                the pre-allocated buffer that will hold the read data
                (numpy, cupy, torch.Tensor arrays)

            start
                variable offset dimensions

            count
                variable local dimensions from offset

            block_id
                (int) Required for reading local variables, local array, and local
                value.

            step_selection
                (list): On the form of [start, count].

            defer_read
                False: read now and blocking wait for completion (Sync mode)
                True: defer reading all requests until read_complete().
                      The returned numpy array will be filled with data
                      only after calling read_complete().
        """
        variable = self._io.inquire_variable(name)
        if not variable:
            raise ValueError()

        self.read_in_buffer(
            variable, buffer, start, count, block_id, step_selection, defer_read=defer_read
        )

    @singledispatchmethod
    def read(
        self,
        variable: Variable,
        start=[],
        count=[],
        block_id=None,
        step_selection=None,
        defer_read: bool = False,
    ):
        """
        Read a variable.
        Random access read allowed to select steps.

        Parameters
            variable
                adios2.Variable object to be read
                Use variable.set_selection(), set_block_selection(), set_step_selection()
                to prepare a read

            start
                variable offset dimensions

            count
                variable local dimensions from offset

            block_id
                (int) Required for reading local variables, local array, and local
                value.

            step_selection
                (list): On the form of [start, count].

            defer_read
                False: read now and blocking wait for completion (Sync mode)
                True: defer reading all requests until read_complete().
                        The returned numpy array will be filled with data
                        only after calling read_complete().
        Returns
            array
                resulting array from selection
        """

        variable = self._set_variable_settings(variable, start, count, block_id, step_selection)
        if variable.type() == "string" and variable.single_value() is True:
            return self._engine.get(variable)
        return self._read_var(variable, defer_read=defer_read)

    @read.register(str)
    def _(
        self,
        name: str,
        start=[],
        count=[],
        block_id=None,
        step_selection=None,
        defer_read: bool = False,
    ):
        """
        Read a variable.
        Random access read allowed to select steps.

        Parameters
            name
                variable to be read

            start
                variable offset dimensions

            count
                variable local dimensions from offset

            block_id
                (int) Required for reading local variables, local array, and local
                value.

            step_selection
                (list): On the form of [start, count].

            defer_read
                False: read now and blocking wait for completion (Sync mode)
                True: defer reading all requests until read_complete().
                        The returned numpy array will be filled with data
                        only after calling read_complete().
        Returns
            array
                resulting array from selection
        """
        variable = self._io.inquire_variable(name)
        if not variable:
            raise ValueError()

        return self.read(variable, start, count, block_id, step_selection, defer_read=defer_read)

    def read_complete(self):
        """
        Complete reading all deferred read requests.
        The returned numpy arrays of each read(..., defer_read=True) will be
        filled with data after this call.
        """
        self._engine.perform_gets()

    def write_attribute(self, name, content, variable_name="", separator="/"):
        """
        writes a self-describing single value array (numpy) variable

        Parameters
            name
                attribute name

            array
                attribute numpy array data

            variable_name
                if attribute is associated with a variable

            separator
                concatenation string between variable_name and attribute
                e.g. variable_name + separator + name ("var/attr")
                Not used if variable_name is empty
        """
        attribute = self._io.inquire_attribute(name, variable_name, separator)
        if not attribute:
            attribute = self._io.define_attribute(name, content, variable_name, separator)

    def read_attribute(self, name, variable_name="", separator="/"):
        """
        Reads a numpy based attribute

        Parameters
            name
                attribute name

            variable_name
                if attribute is associated with a variable

            separator
                concatenation string between variable_name and attribute
                e.g. variable_name + separator + name (var/attr)
                Not used if variable_name is empty

        Returns
            array
                resulting array attribute data
        """
        attribute = self._io.inquire_attribute(name, variable_name, separator)
        if not attribute:
            raise KeyError()

        if attribute.type() == "string":
            return attribute.data_string()

        return attribute.data()

    def read_attribute_string(self, name, variable_name="", separator="/"):
        """
        Reads a numpy based attribute

        Parameters
            name
                attribute name

            variable_name
                if attribute is associated with a variable

            separator
                concatenation string between variable_name and attribute
                e.g. variable_name + separator + name (var/attr)
                Not used if variable_name is empty

        Returns
            array
                resulting array attribute data
        """
        attribute = self._io.inquire_attribute(name, variable_name, separator)
        if not attribute:
            raise KeyError()

        return attribute.data_string()

    def begin_step(self, *, timeout=DEFAULT_TIMEOUT_SEC):
        """
        Write mode: declare the starting of an output step. Pass data in
        stream.write() and stream.write_attribute(). All data will be published
        in end_step().

        Read mode: in streaming mode releases the current step (no effect
        in file based engines)
        """
        if self._read_mode:
            mode = bindings.StepMode.Read
        else:
            mode = bindings.StepMode.Append

        if not self._engine.between_step_pairs():
            return self._engine.begin_step(mode=mode, timeoutSeconds=timeout)
        return bindings.StepStatus.OtherError

    def end_step(self):
        """
        Write mode: declaring the end of an output step. All data passed in
        stream.write() and all attributes passed in stream.write_attribute()
        will be published for consumers.

        Read mode: in streaming mode releases the current step (no effect
        in file based engines)
        """
        self._engine.end_step()

    def close(self):
        """
        Closes stream, thus becoming unreachable.
        Not required if using open in a with-as statement.
        Required in all other cases per-open to avoid resource leaks.
        """
        self._engine.close()
        self._engine = None
        if not self._read_mode:
            self._io.flush_all()
            self._adios.flush_all()
        self._io = None
        self._adios = None

    def current_step(self):
        """
        Inspect current step of the stream. The steps run from 0.

        Note that in a real stream, steps from the producer may be missed if
        the consumer is slow and the producer is told to discard steps
        when no one is reading them in time. You may see non-consecutive
        numbers from this function call in this case.

        Use loop_index() to get a loop counter in a for ... .steps() loop.

        Returns
            current step
        """
        return self._engine.current_step()

    def loop_index(self):
        """
        Inspect the loop counter when using for-in loops. This function returns
        consecutive numbers from 0.

        Returns
            the loop counter
        """
        return self.index

    def steps(self, num_steps=0, *, timeout=DEFAULT_TIMEOUT_SEC):
        """
        Returns an interator that can be use to itererate throught the steps.
        In each iteration begin_step() and end_step() will be internally called.

        Write Mode: num_steps is a mandatory argument and should specify the number
        of steps.

        Read Mode: if num_steps is not specified there will be as much iterations
        as provided by the actual engine. If num_steps is given and there is not
        that many steps in a file/stream, an error will occur.

        IMPORTANT NOTE: Do not use with ReadRandomAccess mode.
        """
        if not self._read_mode and num_steps == 0:
            raise RuntimeError("Stream.steps() in write mode requires num_steps")

        if self._mode == bindings.Mode.ReadRandomAccess:
            raise RuntimeError("Stream.steps() is not allowed in ReadRandomAccess mode")

        if self._mode == bindings.Mode.Read and not self.index < 0:
            raise RuntimeError(
                "Stream.steps() can only be called once in Read mode."
                " Close the stream and reopen to run another iteration loop."
            )

        if num_steps > 0:
            self.max_steps = num_steps
        else:
            self.max_steps = maxsize  # engine steps will limit the loop

        self._step_timeout_sec = timeout

        # in write mode we can run yet another loop
        self.index = -1

        return self

    def num_steps(self):
        """
        READ MODE ONLY. Return the number of steps available.
        Note that this is the steps of a file/stream. Each variable has
        its own steps, which needs to inspected with var=stream.inquire_variable() and then
        with var.steps()
        """
        if not self._read_mode:
            raise RuntimeError("num_steps requires Read/ReadRandomAccess mode")

        return self._engine.steps()

    def all_blocks_info(self, name):
        """
        Extracts all available blocks information for a particular variable. This can be
        an expensive function, memory scales up with metadata: steps and blocks per step

        Args:
            name (str): variable name

        Returns:
            list of dictionaries with information of each step
        """
        if not self._read_mode:
            raise RuntimeError("all_blocks_info requires Read/ReadRandomAccess mode")

        if self._mode == bindings.Mode.Read:
            self.begin_step()

        return self._engine.all_blocks_info(name)

    def minmax(self, name, step: Optional[int] = None, block_info_list: list = []) -> tuple:
        """
        Get min/max value for variable:
        - 'r' mode: min/max in the current step
            - Setting 'step' input will cause error.
        - 'rra' mode:
            - the global all-step min/max if 'step' is not given,
            - the given step's min/max

        If 'bis' is given as the result of previous all_blocks_info() call, it will be used
        instead of querying again.

        :param name: Variable name
        :param step: Which step. It will cause error in 'r' mode if set.
        :param bis: Result of previous all_blocks_info() call.

        Return a tuple of min/max values
        """

        if self._mode == bindings.Mode.Read and step is not None:
            raise ValueError("adios2.Stream.minmax() in 'r' mode must not set step")

        variable = self._io.inquire_variable(name)
        if variable is None:
            return None, None
        if variable.type() == "string":
            return "", ""

        bis = []
        if len(block_info_list) == 0:
            if step is not None:
                bis.append(self._engine.blocks_info(name, step))
            else:
                bis = self._engine.all_blocks_info(name)
        else:
            bis = block_info_list

        dtype = type_adios_to_numpy(variable.type())

        if step is not None:
            mins = [dtype(bi["Min"]) for bi in bis[0]]
            maxs = [dtype(bi["Max"]) for bi in bis[0]]
        else:
            mins = [dtype(bi["Min"]) for n in range(len(bis)) for bi in bis[n]]
            maxs = [dtype(bi["Max"]) for n in range(len(bis)) for bi in bis[n]]
        return min(mins), max(maxs)



# SPDX-FileCopyrightText: 2026 Oak Ridge National Laboratory and Contributors
#
# SPDX-License-Identifier: Apache-2.0

import numpy as np
from os import fstat
from .utils import *


def ReadEncodedStringFromBuffer(buf, pos, ID, limit, lenbytes=2):
    # 'lenbytes' bytes length + string without \0
    if lenbytes == 1:
        dt = np.dtype(np.uint8)
    else:
        dt = np.dtype(np.uint16)
    namelen = np.frombuffer(buf, dtype=dt, count=1, offset=pos)[0]
    pos = pos + lenbytes
    if namelen > limit - lenbytes:
        print(
            "ERROR: "
            + ID
            + " string length ({0}) is longer than the limit to stay inside the block ({1})".format(
                namelen, limit - lenbytes
            )
        )
        return False, "", namelen, pos
    name = buf[pos : pos + namelen].decode("ascii")
    pos = pos + namelen
    return True, name, namelen, pos


def ReadEncodedStringArrayFromBuffer(buf, pos, ID, limit, nStrings):
    s = []
    for i in range(nStrings):
        # 2 bytes length + string without \0
        namelen = np.frombuffer(buf, dtype=np.uint16, count=1, offset=pos)[0]
        pos = pos + 2
        if namelen > limit - 2:
            print(
                "ERROR: "
                + ID
                + " string length ({0}) is longer than the "
                "limit to stay inside the block ({1})".format(namelen, limit - 2)
            )
            return False, s, pos
        name = buf[pos : pos + namelen].decode("ascii")
        pos = pos + namelen
        limit = limit - namelen - 2
        s.append(name)
    return True, s, pos


def ReadDimensionCharacteristics(buf, pos):
    ndim = np.frombuffer(buf, dtype=np.uint8, count=1, offset=pos)[0]
    pos = pos + 1
    lgo = np.zeros(ndim, dtype=np.uint64)
    dimLen = np.frombuffer(buf, dtype=np.uint16, count=1, offset=pos)[0]
    pos = pos + 2
    if dimLen != 24 * ndim:
        print(
            "ERROR: Encoded dimension length expected size = {0} bytes, but found {1} bytes".format(
                24 * ndim, dimLen
            )
        )
        return False, pos, ndim, lgo

    lgo = np.frombuffer(buf, dtype=np.uint64, count=3 * ndim, offset=pos)
    pos = pos + 24 * ndim
    return True, pos, ndim, lgo


def bDataToNumpyArray(cData, typeName, nElements, startPos=0):
    if typeName == "byte":
        return np.frombuffer(cData, dtype=np.int8, count=nElements, offset=startPos)
    elif typeName == "char":
        return np.frombuffer(cData, dtype=np.uint8, count=nElements, offset=startPos)
    elif typeName == "short":
        return np.frombuffer(cData, dtype=np.int16, count=nElements, offset=startPos)
    elif typeName == "integer":
        return np.frombuffer(cData, dtype=np.int32, count=nElements, offset=startPos)
    elif typeName == "long":
        return np.frombuffer(cData, dtype=np.int64, count=nElements, offset=startPos)

    elif typeName == "unsigned_byte":
        return np.frombuffer(cData, dtype=np.uint8, count=nElements, offset=startPos)
    elif typeName == "unsigned_short":
        return np.frombuffer(cData, dtype=np.uint16, count=nElements, offset=startPos)
    elif typeName == "unsigned_integer":
        return np.frombuffer(cData, dtype=np.uint32, count=nElements, offset=startPos)
    elif typeName == "unsigned_long":
        return np.frombuffer(cData, dtype=np.uint64, count=nElements, offset=startPos)

    elif typeName == "real":
        return np.frombuffer(cData, dtype=np.float32, count=nElements, offset=startPos)
    elif typeName == "double":
        return np.frombuffer(cData, dtype=np.float64, count=nElements, offset=startPos)
    elif typeName == "long_double":
        return np.frombuffer(cData, dtype=np.float128, count=nElements, offset=startPos)

    elif typeName == "complex":
        return np.frombuffer(cData, dtype=np.complex64, count=nElements, offset=startPos)
    elif typeName == "double_complex":
        return np.frombuffer(cData, dtype=np.complex128, count=nElements, offset=startPos)

    else:
        return np.zeros(1, dtype=np.uint32)


def ReadCharacteristicsFromMetaData(buf, idx, pos, limit, typeID, fileOffset, isVarCharacteristics):
    cStartPosition = pos
    dataTypeName = GetTypeName(typeID)
    print("        Block {0}: ".format(idx))
    print("            Starting offset : {0}".format(fileOffset))
    # 1 byte NCharacteristics
    nChars = np.frombuffer(buf, dtype=np.uint8, count=1, offset=pos)[0]
    pos = pos + 1
    print("            # of Characteristics    : {0}".format(nChars))
    # 4 bytes length
    charLen = np.frombuffer(buf, dtype=np.uint8, count=32, offset=pos)[0]
    pos = pos + 4
    print("            Characteristics Length  : {0}".format(charLen))

    # For attributes, we need to remember the dimensions and size
    # when reading the value
    ndim = 0
    nElems = 1

    for i in range(nChars):
        print("            Characteristics[{0}]".format(i))
        # 1 byte TYPE
        cID = np.frombuffer(buf, dtype=np.uint8, count=1, offset=pos)[0]
        pos = pos + 1
        cName = GetCharacteristicName(cID)
        print("                Type           : {0} ({1}) ".format(cName, cID))
        cLen = GetCharacteristicDataLength(cID, typeID)

        if cName == "dimensions":
            status, pos, ndim, lgo = ReadDimensionCharacteristics(buf, pos)
            if not status:
                return status, pos
            print("                # of Dims      : {0}".format(ndim))
            if ndim > 0:
                print("                Dims (lgo)     : (", end="")
                for d in range(ndim):
                    p = 3 * d
                    nElems = int(nElems * lgo[p])  # need for value later
                    print("{0}:{1}:{2}".format(lgo[p], lgo[p + 1], lgo[p + 2]), end="")
                    if d < ndim - 1:
                        print(", ", end="")
                    else:
                        print(")")

        elif cName == "value" or cName == "min" or cName == "max":
            if dataTypeName == "string":
                namelimit = limit - (pos - cStartPosition)
                status, s, sLen, pos = ReadEncodedStringFromBuffer(
                    buf, pos, "String Value", namelimit
                )
                if not status:
                    return False, pos
                print("                Value          : '" + s + "' ({0} bytes)".format(sLen))
            elif dataTypeName == "string_array":
                namelimit = limit - (pos - cStartPosition)
                status, strList, pos = ReadEncodedStringArrayFromBuffer(
                    buf, pos, "String Array", namelimit, lgo[0]
                )
                if not status:
                    return False, pos
                print("                Value          : [", end="")
                for j in range(len(strList)):
                    print("'" + strList[j] + "'", end="")
                    if j < len(strList) - 1:
                        print(", ", end="")
                print("]")

            else:
                if isVarCharacteristics:
                    cData = buf[pos : pos + cLen]
                    pos = pos + cLen
                    data = bDataToNumpyArray(cData, dataTypeName, 1)
                    print("                Value          : {0}  ({1} bytes)".format(data[0], cLen))
                else:  # attribute value characteristics are different
                    dataTypeSize = GetTypeSize(typeID)
                    nBytes = int(nElems * dataTypeSize)
                    cData = buf[pos : pos + nBytes]
                    pos = pos + nBytes
                    data = bDataToNumpyArray(cData, dataTypeName, nElems)
                    print("                Value          : [", end="")
                    for j in range(nElems):
                        print("{0}".format(data[j]), end="")
                        if j < nElems - 1:
                            print(", ", end="")
                    print("]")

        elif cName == "offset" or cName == "payload_offset":
            cData = buf[pos : pos + cLen]
            pos = pos + cLen
            data = bDataToNumpyArray(cData, "unsigned_long", 1)
            print("                Value          : {0}  ({1} bytes)".format(data[0], cLen))
        elif cName == "time_index" or cName == "file_index":
            cData = buf[pos : pos + cLen]
            pos = pos + cLen
            data = bDataToNumpyArray(cData, "unsigned_integer", 1)
            print("                Value          : {0}  ({1} bytes)".format(data[0], cLen))
        elif cName == "minmax":
            nBlocks = np.frombuffer(buf, dtype=np.uint16, count=1, offset=pos)[0]
            print("                nBlocks        : {0}".format(nBlocks))
            pos = pos + 2
            bminmax = bDataToNumpyArray(buf, dataTypeName, 2, pos)
            pos = pos + 2 * cLen
            print("                Min/max        : {0} / {1}".format(bminmax[0], bminmax[1]))
            if nBlocks > 1:
                method = np.frombuffer(buf, dtype=np.uint8, count=1, offset=pos)[0]
                pos = pos + 1
                print("                Division method: {0}".format(method))
                blockSize = np.frombuffer(buf, dtype=np.uint64, count=1, offset=pos)[0]
                pos = pos + 8
                print("                Block size     : {0}".format(blockSize))
                div = np.frombuffer(buf, dtype=np.uint16, count=ndim, offset=pos)
                pos = pos + 2 * ndim
                print("                Division vector: (", end="")
                for d in range(ndim):
                    print("{0}".format(div[d]), end="")
                    if d < ndim - 1:
                        print(", ", end="")
                    else:
                        print(")")
                minmax = bDataToNumpyArray(buf, dataTypeName, 2 * nBlocks, pos)
                pos = pos + 2 * nBlocks * cLen
                for i in range(nBlocks):
                    print(
                        "                Min/max        : {0} / {1}".format(
                            minmax[2 * i], minmax[2 * i + 1]
                        )
                    )
        elif cName == "transform_type":
            # Operator name (8 bit length)
            namelimit = limit - (pos - cStartPosition)
            status, s, sLen, pos = ReadEncodedStringFromBuffer(
                buf, pos, "Operator Name", namelimit, lenbytes=1
            )
            if not status:
                return False, pos
            print("                Operator       : '" + s + "' ({0} bytes)".format(sLen))

            # 1 byte TYPE
            typeID = buf[pos]
            pos = pos + 1
            print("                Pre-type       : {0} ({1}) ".format(GetTypeName(typeID), typeID))

            # Pre-transform dimenstions
            status, pos, ndim, lgo = ReadDimensionCharacteristics(buf, pos)
            if not status:
                return status, pos
            print("                Pre-# of dims  : {0}".format(ndim))
            if ndim > 0:
                print("                Pre-Dims (lgo) : (", end="")
                for d in range(ndim):
                    p = 3 * d
                    nElems = int(nElems * lgo[p])  # need for value later
                    print("{0}:{1}:{2}".format(lgo[p], lgo[p + 1], lgo[p + 2]), end="")
                    if d < ndim - 1:
                        print(", ", end="")
                    else:
                        print(")")

            # Operator specific metadata
            omdlen = np.frombuffer(buf, dtype=np.uint16, count=1, offset=pos)[0]
            pos = pos + 2
            print("                Op. data length: {0}".format(omdlen))
            pos = pos + omdlen
        else:
            print(
                "                ERROR: could not understand this "
                "characteristics type '{0}' id {1}".format(cName, cID)
            )
    return True, pos


def ReadPGMD(buf, idx, pos, limit, pgStartOffset):
    # Read one PG index group
    pgStartPosition = pos
    print("    PG {0}: ".format(idx))
    print("        Starting offset : {0}".format(pgStartOffset))

    # 2 bytes PG Length + Name length
    pgLength = np.frombuffer(buf, dtype=np.uint16, count=1, offset=pos)[0]
    pos = pos + 2
    print("        PG length       : {0} bytes (+2 for length)".format(pgLength))
    if pgStartPosition + pgLength + 2 > limit:
        print(
            "ERROR: There is not enough bytes {0} left in PG index block "
            "to read this single PG index ({1} bytes)"
        ).format(limit - pgStartPosition, pgLength)
        return False, pos

    pgNameLen = np.frombuffer(buf, dtype=np.uint16, count=1, offset=pos)[0]
    pos = pos + 2
    if pgStartPosition + pgNameLen > limit:
        print(
            "ERROR: There is not enough bytes {0} left in PG index block "
            "to read the name of this single PG index ({1} bytes)"
        ).format(limit - pos + 2, pgNameLen)
        return False, pos

    pgName = buf[pos : pos + pgNameLen].decode("ascii")
    pos = pos + pgNameLen
    print("        PG Name         : '" + pgName + "' ({0} bytes)".format(pgNameLen))

    # ColumnMajor (host language Fortran) 1 byte, 'y' or 'n'
    isColumnMajor = buf[pos]  # this is an integer value
    pos = pos + 1
    if isColumnMajor != ord("y") and isColumnMajor != ord("n"):
        print(
            "ERROR: Next byte for isColumnMajor must be 'y' or 'n' "
            "but it isn't = {0} (={1})".format(chr(isColumnMajor), isColumnMajor)
        )
        return False
    print("        isColumnMajor   : " + chr(isColumnMajor))

    processID = np.frombuffer(buf, dtype=np.uint32, count=1, offset=pos)[0]
    pos = pos + 4
    print("        process ID      : {0}".format(processID))

    pgTimeNameLen = np.frombuffer(buf, dtype=np.uint16, count=1, offset=pos)[0]
    pos = pos + 2
    if pgStartPosition + pgTimeNameLen > limit:
        print(
            "ERROR: There is not enough bytes {0} left in PG index block "
            "to read the name of this single PG index ({1} bytes)"
        ).format(limit - pos + 2, pgTimeNameLen)
        return False, pos

    pgTimeName = buf[pos : pos + pgTimeNameLen].decode("ascii")
    pos = pos + pgTimeNameLen
    print("        Timestep Name   : '" + pgTimeName + "' ({0} bytes)".format(pgTimeNameLen))

    step = np.frombuffer(buf, dtype=np.uint32, count=1, offset=pos)[0]
    pos = pos + 4
    print("        Step            : {0}".format(step))

    ptr = np.frombuffer(buf, dtype=np.uint64, count=1, offset=pos)[0]
    pos = pos + 8
    print("        Offset in data  : {0}".format(ptr))

    return True, pos


def ReadVarMD(buf, idx, pos, limit, varStartOffset):
    # Read one VAR index group
    varStartPosition = pos
    print("    Var {0}: ".format(idx))
    print("        Starting offset : {0}".format(varStartOffset))

    # 4 bytes VAR Index Length
    varLength = np.frombuffer(buf, dtype=np.uint32, count=1, offset=pos)[0]
    pos = pos + 4
    print("        Var idx length  : {0} bytes (+4 for idx length)".format(varLength))
    if varStartPosition + varLength + 4 > limit:
        print("ERROR: There is not enough bytes in Var index block to read this single Var index")
        return False, pos

    memberID = np.frombuffer(buf, dtype=np.uint32, count=1, offset=pos)[0]
    pos = pos + 4
    print("        MemberID        : {0}".format(memberID))

    namelimit = limit - (pos - varStartPosition)
    status, grpName, grpNameLen, pos = ReadEncodedStringFromBuffer(
        buf, pos, "Group Name", namelimit
    )
    if not status:
        return False, pos
    print("        Group Name      : '" + grpName + "' ({0} bytes)".format(grpNameLen))

    namelimit = limit - (pos - varStartPosition)
    status, varName, varNameLen, pos = ReadEncodedStringFromBuffer(buf, pos, "Var Name", namelimit)
    if not status:
        return False, pos
    print("        Var Name        : '" + varName + "' ({0} bytes)".format(varNameLen))

    # print("        Current offset : {0}".format(varStartOffset + pos))
    # namelimit = limit - (pos - varStartPosition)
    # status, varPath, varPathLen, pos = ReadEncodedStringFromBuffer(
    #     buf, pos, "Var Path", namelimit)
    # if not status:
    #     return False, pos
    # print("        Var Path        : '" + varPath +
    #       "' ({0} bytes)".format(varPathLen))

    # 1 byte ORDER (K, C, F)
    order = buf[pos]  # this is an integer value
    pos = pos + 1
    if order != ord("K") and order != ord("C") and order != ord("F"):
        print(
            "ERROR: Next byte for Order must be 'K', 'C', or 'F' but it isn't = {0} (={1})".format(
                chr(order), order
            )
        )
        return False
    print("        Order           : " + chr(order))

    # 1 byte UNUSED
    unused = buf[pos]  # this is an integer value
    pos = pos + 1
    print("        Unused byte     : {0}".format(unused))

    # 1 byte TYPE
    typeID = buf[pos]
    pos = pos + 1
    print("        Type            : {0} ({1}) ".format(GetTypeName(typeID), typeID))

    # 8 byte Number of Characteristics Sets
    cSets = np.frombuffer(buf, dtype=np.uint64, count=1, offset=pos)[0]
    pos = pos + 8
    print("        # of blocks     : {0}".format(cSets))

    #   This loop only reads the number of reported blocks
    #     for i in range(cSets):
    #         # one characteristics block
    #         newlimit = limit - (pos - varStartPosition)
    #         fileOffset = varStartOffset + (pos - varStartPosition)
    #         status, pos = ReadCharacteristicsFromMetaData(
    #             buf, i, pos, newlimit, typeID, fileOffset, True)
    #         if not status:
    #             return False

    #   This loop reads blocks until the reported length of variable index length
    i = 0
    while pos < varStartPosition + varLength:
        # one characteristics block
        newlimit = limit - (pos - varStartPosition)
        fileOffset = varStartOffset + (pos - varStartPosition)
        status, pos = ReadCharacteristicsFromMetaData(
            buf, i, pos, newlimit, typeID, fileOffset, True
        )
        if not status:
            return False
        i = i + 1

    if i != cSets:
        print("ERROR: reported # of blocks (={0}) != # of encoded blocks  (={1})".format(cSets, i))

    return True, pos


def ReadAttrMD(buf, idx, pos, limit, attrStartOffset):
    # Read one ATTR index group
    attrStartPosition = pos
    print("    Attr {0}: ".format(idx))
    print("        Starting offset : {0}".format(attrStartOffset))

    # 4 bytes ATTR Index Length
    attrLength = np.frombuffer(buf, dtype=np.uint32, count=1, offset=pos)[0]
    pos = pos + 4
    print("        Attr idx length : {0} bytes (+4 for idx length)".format(attrLength))
    if attrStartPosition + attrLength + 4 > limit:
        print("ERROR: There is not enough bytes in Attr index block to read this single Attr index")
        return False, pos

    memberID = np.frombuffer(buf, dtype=np.uint32, count=1, offset=pos)[0]
    pos = pos + 4
    print("        MemberID        : {0}".format(memberID))

    namelimit = limit - (pos - attrStartPosition)
    status, grpName, grpNameLen, pos = ReadEncodedStringFromBuffer(
        buf, pos, "Group Name", namelimit
    )
    if not status:
        return False, pos
    print("        Group Name      : '" + grpName + "' ({0} bytes)".format(grpNameLen))

    namelimit = limit - (pos - attrStartPosition)
    status, attrName, attrNameLen, pos = ReadEncodedStringFromBuffer(
        buf, pos, "Attr Name", namelimit
    )
    if not status:
        return False, pos
    print("        Attr Name       : '" + attrName + "' ({0} bytes)".format(attrNameLen))

    # print("        Current offset : {0}".format(attrStartOffset + pos))
    namelimit = limit - (pos - attrStartPosition)
    status, attrPath, attrPathLen, pos = ReadEncodedStringFromBuffer(
        buf, pos, "Attr Path", namelimit
    )
    if not status:
        return False, pos
    print("        Attr Path       : '" + attrPath + "' ({0} bytes)".format(attrPathLen))

    # 1 byte TYPE
    typeID = buf[pos]
    pos = pos + 1
    print("        Type            : {0} ({1}) ".format(GetTypeName(typeID), typeID))

    # 8 byte Number of Characteristics Sets
    cSets = np.frombuffer(buf, dtype=np.uint64, count=1, offset=pos)[0]
    pos = pos + 8
    print("        # of blocks     : {0}".format(cSets))

    for i in range(cSets):
        # one characteristics block
        newlimit = limit - (pos - attrStartPosition)
        fileOffset = attrStartOffset + (pos - attrStartPosition)
        status, pos = ReadCharacteristicsFromMetaData(
            buf, i, pos, newlimit, typeID, fileOffset, False
        )
        if not status:
            return False

    return True, pos


def ReadMetadataStep(f, fileSize, step):
    # Read metadata of one step
    mdStartPosition = f.tell()
    if step > 0:
        print("========================================================")
    print("Step {0}: ".format(step))
    print("    PG Index offset   : {0}".format(mdStartPosition))

    # Read the PG Index

    # 8 bytes PG Count + Index Length
    pgInfo = np.fromfile(f, dtype=np.uint64, count=2)
    pgCount = pgInfo[0]
    pgLength = pgInfo[1]
    print("    # of PGs          : {0}".format(pgCount))
    print("    PG Index length   : {0} bytes (+16 for count+len)".format(pgLength))

    pgStartPosition = f.tell()
    if pgStartPosition + pgLength > fileSize:
        print("ERROR: There is not enough bytes in file to read the PG index")
        return False

    pgmd = f.read(pgLength)
    pgmdPos = 0
    for i in range(pgCount):
        # VMD block
        status, pgmdPos = ReadPGMD(pgmd, i, pgmdPos, pgLength, pgStartPosition + pgmdPos)
        if not status:
            return False

    # Read the VAR Index

    print("    ----------------------------------")
    print("    Var Index offset  : {0}".format(f.tell()))

    # 4 bytes VAR Count + 8 bytes Var Index Length
    varCount = np.fromfile(f, dtype=np.uint32, count=1)[0]
    varLength = np.fromfile(f, dtype=np.uint64, count=1)[0]
    print("    # of Variables    : {0}".format(varCount))
    print("    Var Index length  : {0} bytes (+12 for count+len)".format(varLength))

    varsStartPosition = f.tell()
    if varsStartPosition + varLength > fileSize:
        print("ERROR: There is not enough bytes in file to read the VAR index")
        return False

    varmd = f.read(varLength)
    varmdPos = 0
    for i in range(varCount):
        # VMD block
        status, varmdPos = ReadVarMD(varmd, i, varmdPos, varLength, varsStartPosition + varmdPos)
        if not status:
            return False

    # Read the ATTR Index

    print("    ----------------------------------")
    print("    Attr Index offset : {0}".format(f.tell()))

    # 4 bytes ATTR Count + 8 bytes Attr Index Length
    attrCount = np.fromfile(f, dtype=np.uint32, count=1)[0]
    attrLength = np.fromfile(f, dtype=np.uint64, count=1)[0]
    print("    # of attriables    : {0}".format(attrCount))
    print("    Attr Index length  : {0} bytes (+12 for count+len)".format(attrLength))

    attrsStartPosition = f.tell()
    if attrsStartPosition + attrLength > fileSize:
        print("ERROR: There is not enough bytes in file to read the Attribute index")
        return False

    attrmd = f.read(attrLength)
    attrmdPos = 0
    for i in range(attrCount):
        # VMD block
        status, attrmdPos = ReadAttrMD(
            attrmd, i, attrmdPos, attrLength, attrsStartPosition + attrmdPos
        )
        if not status:
            return False

    return True


def DumpMetaData(fileName):
    print("========================================================")
    print("    Metadata File: " + fileName)
    print("========================================================")
    with open(fileName, "rb") as f:
        fileSize = fstat(f.fileno()).st_size
        status = ReadHeader(f, fileSize, "Metadata")
        step = 0
        while f.tell() < fileSize - 12 and status:
            status = ReadMetadataStep(f, fileSize, step)
            step = step + 1
    return status


if __name__ == "__main__":
    print("ERROR: Utility main program is bp4dbg.py")



# SPDX-FileCopyrightText: 2026 Oak Ridge National Laboratory and Contributors
#
# SPDX-License-Identifier: Apache-2.0

"""ADIOS2 Engine Python wrapper."""

import numpy as np

try:
    import cupy as cp

    ADIOS2_HAS_CUPY = True
except ImportError:
    ADIOS2_HAS_CUPY = False
try:
    import torch

    ADIOS2_HAS_TORCH = True
except ImportError:
    ADIOS2_HAS_TORCH = False

from adios2 import bindings


class Engine:
    """High level representation of the Engine class in the adios2.bindings"""

    def __init__(self, implementation):
        self.impl = implementation

    @property
    def impl(self):
        """Bindings implementation of the class"""
        return self._impl

    @impl.setter
    def impl(self, implementation):
        self._impl = implementation

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close()

    def close(self, transport_index=-1):
        """
        Close an transports of the engine

        Args:
            transportIndex (int): if -1 close all transports
        """
        self.impl.Close(transport_index)

    def steps(self):
        """Returns number of available steps"""
        return self.impl.Steps()

    def current_step(self):
        """Returns the current step"""
        return self.impl.CurrentStep()

    def begin_step(self, *args, **kwargs):
        """Start step"""
        return self.impl.BeginStep(*args, **kwargs)

    def end_step(self):
        """End step"""
        self.impl.EndStep()

    def between_step_pairs(self):
        """End step"""
        return self.impl.BetweenStepPairs()

    def all_blocks_info(self, name):
        """
        Returns a list of BlocksInfo for a all steps

        Args:
            name (str): variable name

        Returns:
            list of BlockInfos
        """
        output = []
        try:
            for step in range(0, self.steps()):
                output.append(self.blocks_info(name, step))
        except ValueError:  # Campaign engine has no steps()
            vstep = 0
            while True:
                try:
                    binfo = self.blocks_info(name, vstep)
                    if len(binfo) == 0:
                        break
                    output.append(binfo)
                    vstep += 1
                except ValueError:
                    break
        return output

    def blocks_info(self, name, step):
        """
        Returns a BlocksInfo for a given step

        Args:
            name (str): variable name
            step (int): step in question

        Returns:
            list of dicts describing each BlockInfo
        """
        return self.impl.BlocksInfo(name, step)

    def put(self, variable, content, mode=bindings.Mode.Deferred):
        """
        Puts content in a variable

        Parameters
            name
                variable name

            content
                variable data values

            mode
                when to perform the communication, (Deferred or Asynchronous).
        """
        if isinstance(content, np.ndarray):
            self.impl.Put(variable.impl, content, mode)
        elif isinstance(content, str):
            self.impl.Put(variable.impl, content)
        elif isinstance(content, list):
            self.impl.Put(variable.impl, content)
        elif not hasattr(content, "__len__"):
            content = np.array([content])
            self.impl.Put(variable.impl, content)
        else:
            if ADIOS2_HAS_CUPY:
                if isinstance(content, cp.ndarray):
                    self.impl.Put(variable.impl, content.data.ptr, mode)
                    return
            if ADIOS2_HAS_TORCH:
                if isinstance(content, torch.Tensor):
                    self.impl.Put(variable.impl, content.data_ptr(), mode)
                    return
            raise ValueError

    def perform_puts(self):
        """Perform the puts calls"""
        self.impl.PerformPuts()

    def perform_data_write(self):
        """Perform the transport data writes"""
        self.impl.PerformDataWrite()

    def get(self, variable, content=None, mode=bindings.Mode.Sync):
        """
        Gets the content of a variable

        Parameters
            name
                variable name

            content
                output variable data values

            mode
                when to perform the communication, (Deferred or Asynchronous).
        Returns
            Content of the variable when the content argument is omitted.
        """
        if isinstance(content, np.ndarray):
            self.impl.Get(variable.impl, content, mode)
            return None

        if ADIOS2_HAS_CUPY:
            if isinstance(content, cp.ndarray):
                self.impl.Get(variable.impl, content.data.ptr, mode)
                return None
        if ADIOS2_HAS_TORCH:
            if isinstance(content, torch.Tensor):
                self.impl.Get(variable.impl, content.data_ptr(), mode)
                return None
        return self.impl.Get(variable.impl, mode)

    def perform_gets(self):
        """Perform the gets calls"""
        self.impl.PerformGets()

    def lock_reader_selections(self):
        """Locks the data selection for read"""
        self.impl.LockReaderSelections()

    def lock_writer_definitions(self):
        """Locks the data selection for write"""
        self.impl.LockWriterDefinitions()

    def flush(self, transport_index=-1):
        """Flush all transports attached to this Engine instance"""
        self.impl.Flush(transport_index)

    def get_metadata(self):
        """
        Get metadata of an open file in a serialized form that can be
        sent to other processes and used to open the file again by
        avoiding reading the metadata from disk.
        """
        return self.impl.GetMetadata()



# SPDX-FileCopyrightText: 2026 Oak Ridge National Laboratory and Contributors
#
# SPDX-License-Identifier: Apache-2.0

LocalValueDim = 18446744073709551613

dataTypes = {
    -1: "unknown",
    0: "byte",
    1: "short",
    2: "integer",
    4: "long",
    50: "unsigned_byte",
    51: "unsigned_short",
    52: "unsigned_integer",
    54: "unsigned_long",
    5: "real",
    6: "double",
    7: "long_double",
    9: "string",
    10: "complex",
    11: "double_complex",
    12: "string_array",
    55: "char",
}

dataTypeSize = {
    -1: 0,
    0: 1,
    1: 2,
    2: 4,
    4: 8,
    50: 1,
    51: 2,
    52: 4,
    54: 8,
    5: 4,
    6: 8,
    7: 16,
    9: 0,
    10: 8,
    11: 16,
    12: 0,
    55: 1,
}


def GetTypeName(typeID):
    name = dataTypes.get(typeID)
    if name is None:
        name = "unknown type"
    return name


def GetTypeSize(typeID):
    size = dataTypeSize.get(typeID)
    if size is None:
        size = 0
    return size


CharacteristicNames = {
    0: "value",
    1: "min",
    2: "max",
    3: "offset",
    4: "dimensions",
    5: "var_id",
    6: "payload_offset",
    7: "file_index",
    8: "time_index",
    9: "bitmap",
    10: "stat",
    11: "transform_type",
    12: "minmax",
}


def GetCharacteristicName(cID):
    name = CharacteristicNames.get(cID)
    if name is None:
        name = "unknown characteristic"
    return name


def GetCharacteristicDataLength(cID, typeID):
    name = CharacteristicNames.get(cID)
    if name == "value" or name == "min" or name == "max" or name == "minmax":
        return dataTypeSize[typeID]
    elif name == "offset" or name == "payload_offset":
        return 8
    elif name == "file_index" or name == "time_index":
        return 4
    else:
        return 0


# Read Header info 64 bytes
# fileType: Data, Metadata, Index Table
def ReadHeader(f, fileSize, fileType, verbose):
    status = True
    if fileSize < 64:
        print("ERROR: Invalid " + fileType + ". File is smaller than the header (64 bytes)")
        return False
    header = f.read(64)
    hStr = header.decode("ascii")

    versionStr = hStr[0:32].replace("\0", " ")
    major = hStr[32]
    minor = hStr[33]
    micro = hStr[34]
    #    unused = hStr[35]

    endianValue = header[36]
    if endianValue == 0:
        endian = "yes"
    elif endianValue == 1:
        endian = " no"
    else:
        print(
            "ERROR: byte 28 must be 0 or 1 to indicate endianness of "
            "the data. It is however {0} in this file".format(endianValue)
        )
        status = False

    bpversion = int(header[37])
    bpminorversion = int(header[38])
    active = int(header[39])
    if active == 0:
        activeStr = " no"
    else:
        activeStr = "yes"
    iscolumnmajor = hStr[40]
    if iscolumnmajor == "n":
        clmnStr = "   no"
    else:
        clmnStr = "  yes"
    # 45..63 unused

    if not verbose:
        return status

    print(
        "---------------------------------------------------------"
        "---------------------------------------------------------"
    )
    print(
        "|        Version string            |Major|Minor|Patch"
        "|unused|Endian|BP version|BP minor|Active|ColumnMajor|unused|"
    )
    print(
        "|          32 bytes                |  1B |  1B |  1B "
        "|  1B  |  1B  |    1B    |   1B   |  1B  |    1B     |  23B |"
    )
    print(
        "+----------------------------------+-----+-----+-----"
        "+------+------+----------+--------+------+-----------+------+"
    )
    print(
        "| {0} | {1} | {2} | {3} |      | {4}  "
        "|    {5}   |  {6}   | {7}  | {8}     |      |".format(
            versionStr,
            str(major).center(3),
            str(minor).center(3),
            str(micro).center(3),
            endian,
            str(bpversion).center(3),
            str(bpminorversion).center(3),
            activeStr,
            clmnStr,
        )
    )
    print(
        "---------------------------------------------------------"
        "---------------------------------------------------------"
    )
    return status


if __name__ == "__main__":
    print("ERROR: Utility main program is bp5dbg.py")
